# Django
from django.urls import reverse_lazy
from django.utils.html import format_html

# Shortcuts
from superadmin.shortcuts import get_slug_or_pk

# Utils
from superadmin.utils import import_class
from superadmin import settings


class BreadcrumbMixin:
    """Clase base que contiene la información común de todas las subclases"""

    def get_menu_in_path(self, path):
        Menu = import_class("superadmin.models", "Menu")
        if not Menu:
            return
        if not path:
            return

        try:
            menu = Menu.objects.get(route=path)
            return menu
        except Menu.DoesNotExist:
            path = "/".join(path.split("/")[0:-1])
            return self.get_menu_in_path(path)

    def get_breadcrumb_text(self, action):
        attr = "breadcrumb_%s_text" % action
        if hasattr(self, "site"):
            text = getattr(self.site, attr, "")
        else:
            text = getattr(settings, attr.upper(), "")

        return format_html(text)

    def get_base(self, menu):
        base = [(menu.name, f"/{menu.route}/")]
        if menu.parent:
            base = self.get_base(menu.parent) + base
        return base

    def get_base_breadcrumbs(self):
        base_breadcrumbs = [(self.get_breadcrumb_text("home"), "/")]

        menu = self.get_menu_in_path(self.request.path[1:-1])
        if not menu:
            return base_breadcrumbs
        base_breadcrumbs.extend(self.get_base(menu))
        return base_breadcrumbs

    def get_create_breadcrumbs(self):
        breadcrumbs = self.get_base_breadcrumbs()
        breadcrumbs.append((self.get_breadcrumb_text(self.action), "#"))
        return breadcrumbs

    def get_update_breadcrumbs(self):
        breadcrumbs = self.get_detail_breadcrumbs()
        breadcrumbs.append((self.get_breadcrumb_text(self.action), "#"))
        return breadcrumbs

    def get_list_breadcrumbs(self):
        return self.get_base_breadcrumbs()

    def get_detail_breadcrumbs(self):
        """Obtiene el breadcumb para Detail View"""
        url_name = self.site.get_url_name("detail")
        breadcrumbs = self.get_list_breadcrumbs()
        breadcrumbs.append(
            (
                self.get_breadcrumb_text("detail") or str(self.object),
                reverse_lazy(
                    url_name,
                    kwargs=get_slug_or_pk(self.object, slug_field=self.site.slug_field),
                ),
            )
        )
        return breadcrumbs

    def get_delete_breadcrumbs(self):
        breadcrumbs = self.get_detail_breadcrumbs()
        breadcrumbs.append((self.get_breadcrumb_text(self.action), "#"))
        return breadcrumbs

    def get_breadcrumbs(self):
        if hasattr(self, "action"):
            attr = getattr(self, "get_%s_breadcrumbs" % self.action)
            return attr()
        return self.get_base_breadcrumbs()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        breadcrumbs = {"breadcrumbs": self.get_breadcrumbs()}

        if "site" in context:
            context["site"].update(breadcrumbs)
        else:
            context.update({"site": breadcrumbs})
        return context



"""Mixins for formsets"""

# Django
from django.shortcuts import redirect
from django.db import transaction
from django.contrib import messages

# Utils
from superadmin.services import FieldService


class FormsetList:
    formsets = dict()

    """
    formsets = {
        "invoice_formset": Formset
    }
    """

    def __init__(self, formsets):
        self.formsets = dict()

        for key, formset_class in formsets.items():
            self.formsets.update(
                {
                    key: {
                        "class": formset_class,
                    }
                }
            )

    def is_valid(self):
        errors = [
            fs["instance"].errors
            for fs in self.formsets.values()
            if not fs["instance"].is_valid()
        ]
        return not errors

    def get_headers(self):
        headers = {
            f"{key}_headers": value["headers"] for key, value in self.formsets.items()
        }
        return headers

    def get_instances(self, **kwargs):
        related_initial = None
        if "initial" in kwargs and "related_initial" in kwargs["initial"]:
            related_initial = kwargs.get("initial").get("related_initial")

        for key in self.formsets:
            formset_class = self.formsets[key]["class"]
            if related_initial:
                kwargs["initial"] = related_initial.get(formset_class.model, [])
            else:
                kwargs["initial"] = []
            instance = formset_class(**kwargs)
            instance.extra += len(kwargs["initial"])

            headers = [
                FieldService.get_field_label(formset_class.form.Meta.model, field.name)
                for field in instance.empty_form.visible_fields()
                if field.name in formset_class.form.Meta.fields
            ]
            self.formsets.get(key).update(
                {
                    "instance": instance,
                    "headers": headers,
                }
            )

        instances = {key: value["instance"] for key, value in self.formsets.items()}
        return instances

    def get_formset(self, name):
        return self.formsets[name]["instance"]


class FormsetMixin:
    """Class for add multiple formsets in form"""

    formsets = dict()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.formsets = FormsetList(formsets=self.formsets)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        formsets = self.get_formsets()
        headers = self.get_headers()
        context.update(**formsets, **headers)
        return context

    def formsets_valid(self, formsets, form):
        with transaction.atomic():
            self.object = form.save()
            for formset in formsets:
                formset.instance = self.object
                formset.save()

        messages.success(self.request, "Se ha guardado correctamente.")
        return redirect(self.get_success_url())

    def formsets_invalid(self, formsets, form):
        for formset in formsets:
            for error in formset.errors:
                form.errors.update(error)
        return super().form_invalid(form)

    def get_headers(self):
        headers = self.formsets.get_headers()
        return headers

    def get_formsets(self):
        """Method to get all formsets"""
        formsets = self.formsets.get_instances(**self.get_form_kwargs())
        return formsets

    def get_formset(self, name):
        return self.formsets.get_formset(name)

    def post(self, request, *args, **kwargs):
        self.object = self.get_object() if self.action == "update" else None
        form = self.get_form()
        formsets = self.get_formsets().values()

        if self.formsets.is_valid() and form.is_valid():
            return self.formsets_valid(formsets, form)
        else:
            return self.formsets_invalid(formsets, form)



# Python
import inspect
from importlib import import_module

# Django
from django.db.models import Q
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.views.generic import View
from django.apps import apps

# Models
from ...models import Action, Menu


def get_views_catalog():
    module_name = "django.views"

    def map_module(module_name):
        try:
            module = import_module(module_name)
        except ModuleNotFoundError:
            return []

        names = [
            name
            for name, cand in inspect.getmembers(module, inspect.isclass)
            if issubclass(cand, View)
        ]
        for name, mod in inspect.getmembers(module, inspect.ismodule):
            names.extend(map_module(f"{module_name}.{name}"))

        return names

    names = set(map_module(module_name))
    return list(names)


def get_actions_and_elements(app_config, VIEWS_CATALOG):
    actions = {
        action.element: action
        for action in Action.objects.filter(app_label=app_config.label)
    }

    model_elements = {
        model._meta.model_name: (
            f"{app_config.verbose_name.capitalize()} | {model._meta.verbose_name.capitalize()}"
        )
        for model in app_config.get_models()
    }

    try:
        module = import_module(f"{app_config.name}.views")
        view_elements = {
            name: (f"{app_config.verbose_name.capitalize()} | {name}")
            for name, candidate in inspect.getmembers(module, inspect.isclass)
            if issubclass(candidate, View) and not candidate.__name__ in VIEWS_CATALOG
        }
    except (ModuleNotFoundError, ImportError):
        view_elements = {}

    elements = {**model_elements, **view_elements}

    return actions, elements, model_elements, view_elements


def create_actions():
    VIEWS_CATALOG = get_views_catalog()

    for app in apps.get_app_configs():
        actions, elements, m, v = get_actions_and_elements(app, VIEWS_CATALOG)

        if not elements:
            continue

        acts = []
        acts += [
            Action(
                to=Action.ToChoices.MODEL,
                app_label=app.label,
                name=name,
                element=element,
            )
            for (element, name) in m.items()
            if element not in actions
        ]
        acts += [
            Action(
                to=Action.ToChoices.CLASSVIEW,
                app_label=app.label,
                name=name,
                element=element,
            )
            for (element, name) in v.items()
            if element not in actions
        ]

        Action.objects.bulk_create(acts)


def build_menu(data):
    create_actions()
    Menu.objects.all().delete()
    default_action = Action.objects.get(app_label="superadmin", element="ModuleView")

    def permissions(data, action):
        content_type = ContentType.objects.get_for_model(Permission)
        for perm in data:
            obj, created = Permission.objects.update_or_create(
                codename=perm.get("codename"),
                content_type=content_type,
                defaults={"name": perm.get("name")},
            )
            action.permissions.add(obj)

    def build(name, data, action, parent=None, menu_list=[]):
        default = action
        is_group = True
        if "app" in data:
            element = data["model"] if "model" in data else data["view"]
            default = Action.objects.get(app_label=data["app"], element=element)
            is_group = False
            if "permissions" in data:
                permissions(data["permissions"], default)

        current = Menu.objects.create(
            parent=parent,
            name=name.capitalize(),
            action=default,
            is_group=is_group,
            sequence=len(menu_list) + 1,
        )
        menu_list.append(current)
        if is_group:
            for key in data:
                build(key, data[key], action, parent=current, menu_list=menu_list)
        return menu_list

    menus = []
    for key in data:
        build(key, data[key], default_action, menu_list=menus)



"""Classes and functios for register site models"""

# Django
from superadmin.services import FieldService
from django.conf import settings
from django.utils.text import slugify
from django.core.exceptions import ImproperlyConfigured
from django.db.models.base import ModelBase
from django.db import connection
from django.urls import path, include
from django.apps import apps

# Local
from .views import FilterView
from .utils import import_mixin


class Site:
    """Site class"""

    def __init__(self, name="site"):
        self._registry = {}
        self.name = name

    def register(self, model, site_class):
        """Registra las clases en el auto site"""
        if isinstance(model, str):
            try:
                app_label, model_name = model.split(".")
                model = apps.get_model(app_label, model_name)
            except ValueError:
                raise Exception(
                    "The model_name passed must be contains app label and model name like 'app_label.model_name'"
                )
        if not isinstance(model, ModelBase):
            raise Exception("The model passed cannot be registered but not is a model.")

        if model._meta.abstract:
            raise ImproperlyConfigured(
                "The model %s is abstract, so it cannot be registered with superadmin."
                % model.__name__
            )

        if model in self._registry:
            raise Exception("The model %s is already registered" % model.__name__)

        self._registry[model] = site_class(model)

    def is_registered(self, model):
        """
        Check if a model class is registered with this `Site`.
        """
        return model in self._registry

    def get_modelsite(self, model):
        if not isinstance(model, ModelBase):
            raise Exception("The model passed is not a Model.")

        if not self.is_registered(model):
            raise Exception("The model %s is not registered" % model.__name__)

        return self._registry[model]

    def get_model_urls(self, menu):
        urlpatterns = []
        model = menu.action.model
        if model and self.is_registered(model):
            model_site = self.get_modelsite(model)
            urlpatterns = [path(f"{menu.route}/", include(model_site.urls))]

        return urlpatterns

    def get_view_urls(self, menu):
        from django.contrib.auth.mixins import PermissionRequiredMixin

        urlpatterns = []
        mixin = import_mixin("BreadcrumbMixin")
        view = menu.action.view
        view.menu = None
        if PermissionRequiredMixin not in view.__bases__:
            view.__bases__ = (PermissionRequiredMixin, mixin, *view.__bases__)
        view.permission_required = menu.action.get_permissions()
        if view:
            urlpatterns = [
                path(
                    route=f"{menu.route}/",
                    view=view.as_view(menu=menu),
                    name=slugify(menu.name),
                )
            ]
        return urlpatterns

    def get_menu_urls(self, menu):
        urlpatterns = []
        if menu.action.to == menu.action.ToChoices.MODEL:
            urlpatterns.extend(self.get_model_urls(menu))
        else:
            urlpatterns.extend(self.get_view_urls(menu))

        return urlpatterns

    def get_menus(self):
        menus = []
        try:
            Menu = apps.get_model("superadmin", "Menu")
            if Menu._meta.db_table in connection.introspection.table_names():
                menus = Menu.objects.select_related("action").all()
        except LookupError as error:
            if settings.DEBUG:
                print(error)
        return menus

    def get_urls(self):
        """Obtiene las urls de auto site"""

        # def wrap(view, cacheable=False):
        #   def wrapper(*args, **kwargs):
        #       return self.admin_view(view, cacheable)(*args, **kwargs)
        #       wrapper.admin_site = self
        #       return update_wrapper(wrapper, view)

        urlpatterns = []
        sites_in_menu = []
        menus = self.get_menus()
        for menu in menus:
            urlpatterns += self.get_menu_urls(menu)
            if self.is_registered(menu.action.model):
                sites_in_menu.append(menu.action.model)
        for model, model_site in self._registry.items():
            if model not in sites_in_menu:
                info = model_site.get_info()
                url_format = "%s/%s/" % info
                urlpatterns += [path(url_format, include(model_site.urls))]
        urlpatterns += [
            path(
                route="filter/<slug:app>/<slug:model>/<slug:field>/",
                view=FilterView.as_view(),
                name="filter",
            )
        ]
        return urlpatterns

    @property
    def urls(self):
        """Permite registrar las URLs en el archivo de urls del proyecto"""
        return self.get_urls(), "site", self.name


site = Site()



# Python
import json

# Django
from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django import template
from django.urls import NoReverseMatch

# Local
from .. import site
from ..shortcuts import get_urls_of_site

register = template.Library()

# Filter utilities
@register.filter("json")
def get_json(value):
    return json.dumps(value)


@register.filter("chunks")
def chunks(value, number):
    number = int(number)
    chunks = [value[i : i + number] for i in range(0, len(value), number)]
    return chunks


@register.filter("subtract")
def subtract(value, number):
    number = float(number)
    return value - number


@register.filter("multiply")
def multiply(value, number):
    number = float(number)
    return value * number


@register.simple_tag()
def site_url(instance, action):
    model_site = site.get_modelsite(instance.__class__)
    urls = get_urls_of_site(model_site, object=instance)
    if action in urls:
        return urls[action]
    raise NoReverseMatch("The action '%s' doesn't exist in the model" % action)


@register.simple_tag()
def has_perm(user, perm):
    return user.has_perm(perm)


@register.simple_tag()
def detail_widget(label, value, widget):
    if widget not in settings.TEMPLATE_WIDGETS_DETAIL:
        if "default" not in settings.TEMPLATE_WIDGETS_DETAIL:
            raise ImproperlyConfigured(
                f"Does not exist template name for '{widget}' or default widget template."
            )
        template_name = settings.TEMPLATE_WIDGETS_DETAIL.get("default")
    else:
        template_name = settings.TEMPLATE_WIDGETS_DETAIL.get(widget)
    return template_name



from .list import ListView
from .create import CreateView
from .update import UpdateView, MassUpdateView
from .detail import DetailView
from .delete import DeleteView, MassDeleteView
from .duplicate import DuplicateView
from .filter import FilterView
from .base import ModuleView



# Python
import operator
from functools import reduce

# Django
from django.http import Http404
from django.db.models import Q

# Local
from ..services import FieldService, FilterService


class FilterMixin:
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        opts = {
            "all_records": self.all_records,
            "filter_fields": self.get_filter_fields(),
            "current_filters": self.get_current_filters(),
        }
        if "site" in context:
            context["site"].update(opts)
        else:
            context.update({"site": opts})
        return context

    def get_params(self, exclude=False):
        DEFAULT = "search"
        EXCLUDE = "__exclude"
        params = {}
        if DEFAULT in self.request.GET:
            params[DEFAULT] = self.request.GET.get(DEFAULT)
            return params
        lookup_params = {
            key: value
            for key, value in self.request.GET.items()
            if self.has_lookup(key)
        }
        for key, value in lookup_params.items():
            lookup = self.has_lookup(key)
            type = FieldService.get_field_type(
                self.site.model, key.split(f"__{lookup}")[0]
            )
            if type == "BooleanField":
                try:
                    value = bool(int(value))
                except ValueError:
                    value = False
            if exclude:
                if EXCLUDE in key:
                    params[key.split(EXCLUDE)[0]] = value
            elif EXCLUDE not in key:
                params[key] = value
        return params

    def get_query(self, exclude=False):
        args = []
        params = self.get_params(exclude=exclude)
        op = operator.__or__
        if "search" in params:
            value = params.get("search")
            for field in self.site.search_fields:
                args.append(Q(**{field: value}))
        else:
            for field, value in params.items():
                args.append(Q(**{field: value}))
            op = operator.__and__
        if args:
            return reduce(op, args)
        return args

    def get_queryset(self):
        queryset = super().get_queryset()
        self.all_records = queryset.count()
        query = self.get_query()
        if query:
            queryset = queryset.filter(query)
        exclude = self.get_query(exclude=True)
        if exclude:
            queryset = queryset.exclude(exclude)
        return queryset

    def get_filter_fields(self):
        fields = []
        for field in self.site.filter_fields:
            field_name = field.split(":")[0]
            fields.append(
                {
                    "name": field_name,
                    "label": FieldService.get_field_label(self.site.model, field),
                    "app_name": self.site.model._meta.app_label,
                    "model_model": self.site.model._meta.model_name,
                }
            )
        return fields

    def get_current_filters(self):
        filters = []
        params = {
            **self.get_params(exclude=True),
            **self.get_params(exclude=False),
        }
        if "search" in params:
            return filters
        for key, value in params.items():
            lookup = self.has_lookup(key)
            lookup_label = FilterService.get_lookup_label(lookup)
            field = key.split(f"__{lookup}")[0]
            for filter_field in self.site.filter_fields:
                if field in filter_field:
                    try:
                        field, field_label = filter_field.split(":")
                    except ValueError:
                        field_label = FieldService.get_field_label(
                            self.site.model, field
                        )
            if not field_label:
                raise Http404
            choices = FilterService.get_choices(self.site.model, field)
            if choices:
                search = int(value) if type(value) == bool else value
                if hasattr(choices, "model"):
                    search_label = str(choices.get(pk=value))
                else:
                    try:
                        search = int(search)
                    except ValueError:
                        pass
                    choices = dict(choices)
                    search_label = choices.get(search)
            else:
                search = value
                search_label = value

            filters.append(
                {
                    "field": field,
                    "field_label": field_label,
                    "lookup": lookup,
                    "lookup_label": lookup_label,
                    "search": search,
                    "search_label": search_label,
                }
            )
        return filters

    def has_lookup(self, field):
        for lookup in FilterService.get_flatten_lookups():
            if lookup in field:
                return lookup



from django.conf import settings

BREADCRUMB_HOME_TEXT = getattr(settings, "BREADCRUMB_HOME_TEXT", "home")
BREADCRUMB_CREATE_TEXT = getattr(settings, "BREADCRUMB_CREATE_TEXT", "create")
BREADCRUMB_UPDATE_TEXT = getattr(settings, "BREADCRUMB_UPDATE_TEXT", "update")
BREADCRUMB_DETAIL_TEXT = getattr(settings, "BREADCRUMB_DETAIL_TEXT", "detail")
BREADCRUMB_DELETE_TEXT = getattr(settings, "BREADCRUMB_DELETE_TEXT", "delete")

BOOLEAN_YES = getattr(settings, "BOOLEAN_YES", "Yes")
BOOLEAN_NO = getattr(settings, "BOOLEAN_NO", "No")

TEMPLATE_WIDGETS = getattr(settings, "TEMPLATE_WIDGETS", {})



""" """
# Django
from django.apps import apps
from django.http import JsonResponse
from django.views.generic import View

# Local
from ..services import FieldService, FilterService


class FilterView(View):
    http_method_names = ["get"]

    def get(self, request, *args, **kwargs):
        model = apps.get_model(self.app_name, self.model_name)
        lookups = self.get_lookups(model, self.field)
        choices = self.get_choices(model, self.field)
        return JsonResponse({"lookups": lookups, "choices": choices})

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.app_name = kwargs.get("app")
        self.model_name = kwargs.get("model")
        self.field = kwargs.get("field")

    def get_lookups(self, model, field):
        lookups = [
            {"id": lookup, "text": label}
            for lookup, label in FilterService.get_field_lookups(model, field)
        ]
        return lookups

    def get_choices(self, model, field):
        choices = FilterService.get_choices(model, field)
        if hasattr(choices, "model"):
            choices = [{"id": obj.id, "text": str(obj)} for obj in choices]
        else:
            choices = [{"id": value, "text": label} for value, label in choices]
        return choices



"""Mixins for autosite"""

# Django
from django.contrib.auth.mixins import (
    PermissionRequiredMixin as DjangoPermissionRequiredMixin,
)


"""
class MultiplePermissionRequiredModuleMixin(DjangoPermissionRequiredMixin):

    def has_permission(self):
        user = self.request.user
        if self.request.user.is_authenticated and self.request.user.is_superuser:
            return True
        permissions = list()
        ctx = self.get_context_data()
        for model in ctx["models_permissions"]:
            permissions.append(f"{model._meta.app_label}.view_{model._meta.model_name}")
            permissions.append(f"{model._meta.app_label}.add_{model._meta.model_name}")
            permissions.append(
                f"{model._meta.app_label}.change_{model._meta.model_name}"
            )
        return any(user.has_perm(permission) for permission in permissions)
"""


class PermissionRequiredMixin(DjangoPermissionRequiredMixin):
    """Verifica los permisos de acceso al modelo"""

    def get_permission_required(self):
        app = self.model._meta.app_label
        model = self.model._meta.model_name

        if self.action == "create":
            permissions = ("add",)
        elif self.action == "update":
            permissions = ("change",)
        elif self.action == "delete":
            permissions = ("delete",)
        else:
            permissions = ("view", "add", "change", "delete")

        perms = (f"{app}.{perm}_{model}" for perm in permissions)
        return perms

    def has_permission(self):
        user = self.request.user
        if all([user.is_authenticated, user.is_superuser, user.is_active]):
            return True
        perms = self.get_permission_required()
        return any(user.has_perm(perm) for perm in perms)



""" """
# Django
from django.views.generic import DetailView as BaseDetailView
from django.contrib.admin.utils import flatten

# Local
from .base import SiteView, get_base_view
from ..services import FieldService
from ..utils import import_all_mixins


class DetailMixin:
    """Definimos la clase que utilizará el modelo"""

    action = "detail"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(self.site.detail_extra_context)

        flatten_results, fieldset_results = self.get_results()
        opts = {
            "results": fieldset_results,
            "flatten_results": flatten_results,
        }

        if "site" in context:
            context["site"].update(opts)
        else:
            context.update({"site": opts})

        return context

    def get_results(self):
        fields = flatten(self.site.detail_fields)
        fields = fields if fields else (field.name for field in self.model._meta.fields)
        results = {}
        for field in fields:
            label = FieldService.get_field_label(self.object, field)
            value = FieldService.get_field_value(self.object, field)
            type = FieldService.get_field_type(self.object, field)
            results[field] = (label, value, type)

        flatten_results = results.values()
        fieldset_results = []
        for fieldset in self.site.detail_fields:
            if isinstance(fieldset, (list, tuple)):
                fieldset_results.append(
                    {
                        "bs_cols": int(12 / len(fieldset)),
                        "fields": [results.get(field, ()) for field in fieldset],
                    }
                )
            else:
                fieldset_results.append(
                    {
                        "bs_cols": 12,
                        "fields": [
                            results.get(fieldset, ()),
                        ],
                    }
                )

        return flatten_results, fieldset_results

    def get_slug_field(self):
        return self.site.slug_field or super().get_slug_field()


class DetailView(SiteView):
    def view(self, request, *args, **kwargs):
        """ Crear la List View del modelo """
        # Class
        mixins = import_all_mixins() + [DetailMixin]
        View = get_base_view(BaseDetailView, mixins, self.site)

        # Set attributes
        View.__bases__ = (*self.site.detail_mixins, *View.__bases__)

        view = View.as_view()
        return view(request, *args, **kwargs)



# Python
import inspect

# Django
from django.conf import settings
from django.views.generic import View
from django.urls import reverse, NoReverseMatch
from django.apps import apps


def get_slug_or_pk(object, slug_field=None):
    res = dict()
    field = slug_field if hasattr(object, slug_field) else "pk"
    if object:
        param = "slug" if hasattr(object, slug_field) else "pk"
        res.update({param: getattr(object, field)})
    return res


def get_object_from_site(site, slug_or_pk):
    model = site.model
    search_field = site.slug_field if hasattr(model, site.slug_field) else "pk"
    try:
        object = model.objects.get(**{search_field: slug_or_pk})
    except model.DoesNotExist:
        object = None
    return object


def get_urls_of_site(site, object=None, user=None):
    urls = {}
    app = site.model._meta.app_label
    model = site.model._meta.model_name
    kwargs = get_slug_or_pk(object, slug_field=site.slug_field)
    for action, perm in (
        ("list", "view"),
        ("create", "add"),
        ("mass_update", "change"),
        ("mass_delete", "delete"),
    ):
        try:
            url_name = site.get_url_name(action)
            if not user:
                urls.update({action: reverse(url_name)})
            elif user.has_perm(f"{app}.{perm}_{model}"):
                urls.update({action: reverse(url_name)})
        except NoReverseMatch:
            if settings.DEBUG:
                print("DEBUG: Url not found: %s" % url_name)
    if not kwargs:
        return urls
    for action, perm in (
        ("update", "change"),
        ("detail", "view"),
        ("delete", "delete"),
        ("duplicate", "add"),
    ):
        try:
            url_name = site.get_url_name(action)
            if not user:
                urls.update({action: reverse(url_name, kwargs=kwargs)})
            elif user.has_perm(f"{app}.{perm}_{model}"):
                urls.update({action: reverse(url_name, kwargs=kwargs)})
        except NoReverseMatch:
            if settings.DEBUG:
                print("DEBUG: Url not found: %s" % url_name)
    return urls



from .formsets import FormsetMixin
from .breadcrumbs import BreadcrumbMixin
from .urls import UrlMixin
from .templates import TemplateMixin
from .permissions import PermissionRequiredMixin
from .filters import FilterMixin



""" Forms for menu """
# Python
import inspect

# Django
from django.forms import BaseModelForm, ChoiceField
from django.forms.models import ModelFormMetaclass as DjangoModelFormMetaclass
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.contrib.admin.utils import flatten
from django.apps import apps

# Models
from .models import Action, Menu

#
from .management.commands.base import get_actions_and_elements


class ModelFormMetaclass(DjangoModelFormMetaclass):
    def __new__(mcs, name, bases, attrs):
        fieldsets = None
        if "Meta" in attrs and hasattr(attrs["Meta"], "fieldsets"):
            fieldsets = attrs["Meta"].fieldsets
            fields = mcs.__fields__(fieldsets)
            if hasattr(attrs["Meta"], "fields"):
                fields = fields + attrs["Meta"].fields
            attrs["Meta"].fields = fields
        new_class = super().__new__(mcs, name, bases, attrs)
        if fieldsets:
            new_class._meta.fieldsets = fieldsets
        return new_class

    def __fields__(fieldsets):
        fields = flatten(fieldsets)
        return tuple(fields)


class ModelForm(BaseModelForm, metaclass=ModelFormMetaclass):
    def get_fieldsets(self):
        sets = list()
        for fieldset in self._meta.fieldsets:
            if isinstance(fieldset, tuple):
                sets.append(
                    {
                        "bs_cols": int(12 / len(fieldset)),
                        "fields": [self[field] for field in fieldset],
                    }
                )
            else:
                sets.append({"bs_cols": 12, "fields": [self[fieldset]]})
        return sets

    def has_fieldsets(self):
        return hasattr(self._meta, "fieldsets")


class BaseActionForm(ModelForm):
    app_label = ChoiceField(label="Nombre de la aplicacióp")
    element = ChoiceField(label="Elemento")

    class Meta:
        model = Action
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        ct = ContentType.objects.get_for_model(Permission)
        codenames = (
            "add_permission",
            "change_permission",
            "delete_permission",
            "view_permission",
        )
        queryset = Permission.objects.filter(content_type=ct).exclude(
            codename__in=codenames
        )
        self.fields["permissions"].queryset = queryset

        APP_CHOICES = (
            (app.label, app.verbose_name.capitalize()) for app in apps.get_app_configs()
        )

        ELEMENT_CHOICES = [
            (element, name)
            for app in apps.get_app_configs()
            for element, name in get_actions_and_elements(app)[1].items()
        ]

        self.fields["app_label"].choices = APP_CHOICES
        self.fields["element"].choices = ELEMENT_CHOICES


class BasePermissionForm(ModelForm):
    class Meta:
        model = Permission
        exclude = ("content_type",)

    def save(self, commit=True):
        perm = super().save(commit=False)
        ct = ContentType.objects.get_for_model(Permission)
        perm.content_type = ct
        if commit:
            perm.save()
        return perm
